/*
 * (C) Copyright 2000
 * Wolfgang Denk, DENX Software Engineering, wd@denx.de.
 *
 * See file CREDITS for list of people who contributed to this
 * project.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.	 See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 * MA 02111-1307 USA
 */

/*
 * Command line user interface to firmware (=U-Boot) environment.
 *
 * Implements:
 *	fw_readenv [ name ... ]
 *		- prints the values of the environment variables
 *		  "name", or the whole environment if no names are
 *		  specified
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <getopt.h>

#include <errno.h>
#include <fcntl.h>
#include <stdio.h>
#include <sys/types.h>   
#include <sys/ioctl.h>
#include <sys/stat.h>
#include <mtd/mtd-user.h>

typedef unsigned char uchar;

#define ENV_SIZE      getenvsize()

typedef struct environment_s {
	ulong crc;			/* CRC32 over data bytes    */
	uchar flags;			/* active or obsolete */
	uchar *data;
} env_t;

static env_t environment;

static uchar active_flag = 1;
static uchar obsolete_flag = 0;

static int flash_io (int mode);
static uchar *envmatch (uchar * s1, uchar * s2);
static int env_init (void);
static int parse_config (void);
static inline ulong getenvsize (void);

char *device = NULL;
int dev_offset = -1;
int env_size = 0;

void
usage(void)
{
	printf("Usage: fw_readenv -n -d device -o offset -s envsize [variables]\n");
}

/*
 * Print the current definition of one, or more, or all
 * environment variables
 */
int
main(int argc, char *argv[])
{
	char *p;
	char *cmdname = *argv;
	uchar *env, *nxt;
	int i;
	int n_flag = 0;
	int ch;
	extern char *optarg;
	extern int optind;


	if ((p = strrchr (cmdname, '/')) != NULL) {
		cmdname = p + 1;
	}

	while ((ch = getopt(argc, argv, "nd:o:s:")) != -1) {
		switch (ch) {
		case 'n':
			n_flag = 1;
			break;
		case 'd':
			device = optarg;
			break;
		case 'o':
			dev_offset = strtol(optarg, (char **)NULL, 16);
			break;
		case 's':
			env_size = strtol(optarg, (char **)NULL, 10);
			break;
		default:
			usage();
			exit(1);
		}
	}
	argc -= optind;
	argv += optind;

	if (device == NULL || dev_offset < 0 || env_size == 0) {
		usage();
		exit(1);
	}

	if (env_init ())
		return;

	if (argc == 0) {		/* Print all env variables  */
		for (env = environment.data; *env; env = nxt + 1) {
			for (nxt = env; *nxt; ++nxt) {
				if (nxt >= &environment.data[ENV_SIZE]) {
					fprintf (stderr, "## Error: "
						"environment not terminated\n");
					return;
				}
			}

			printf ("%s\n", env);
		}
		return;
	}

	for (i = 0; i < argc; ++i) {	/* print single env variables   */
		char *name = argv[i];
		uchar *val = NULL;

		for (env = environment.data; *env; env = nxt + 1) {

			for (nxt = env; *nxt; ++nxt) {
				if (nxt >= &environment.data[ENV_SIZE]) {
					fprintf (stderr, "## Error: "
						"environment not terminated\n");
					return;
				}
			}
			val = envmatch (name, env);
			if (val) {
				if (!n_flag) {
					fputs (name, stdout);
					putc ('=', stdout);
				}
				puts (val);
				break;
			}
		}
		if (!val)
			fprintf (stderr, "## Error: \"%s\" not defined\n", name);
	}
	return (EXIT_FAILURE);
}

static inline ulong getenvsize (void)
{
	ulong rc = env_size - sizeof (long);

	return rc;
}

static int flash_io (int mode)
{
	int fd, fdr, rc, otherdev, len, resid;
	erase_info_t erase;
	char *data = NULL;

	if ((fd = open (device, mode)) < 0) {
		fprintf (stderr,
			"Can't open %s: %s\n",
			device, strerror (errno));
		return (-1);
	}

	len = sizeof (environment.crc);

	if (lseek (fd, dev_offset, SEEK_SET) == -1) {
		fprintf (stderr,
			"seek error on %s: %s\n",
			device, strerror (errno));
		return (-1);
	}
	if (read (fd, &environment, len) != len) {
		fprintf (stderr,
			"CRC read error on %s: %s\n",
			device, strerror (errno));
		return (-1);
	}
	if ((rc = read (fd, environment.data, ENV_SIZE)) != ENV_SIZE) {
		fprintf (stderr,
			"Read error on %s: %s\n",
			device, strerror (errno));
		return (-1);
	}

	if (close (fd)) {
		fprintf (stderr,
			"I/O error on %s: %s\n",
			device, strerror (errno));
		return (-1);
	}

	/* everything ok */
	return (0);
}

/*
 * s1 is either a simple 'name', or a 'name=value' pair.
 * s2 is a 'name=value' pair.
 * If the names match, return the value of s2, else NULL.
 */

static uchar *envmatch (uchar * s1, uchar * s2)
{

	while (*s1 == *s2++)
		if (*s1++ == '=')
			return (s2);
	if (*s1 == '\0' && *(s2 - 1) == '=')
		return (s2);
	return (NULL);
}

/*
 * Prevent confusion if running from erased flash memory
 */
static int env_init (void)
{
	int crc1, crc1_ok;
	uchar *addr1;

	int crc2, crc2_ok;
	uchar flag1, flag2, *addr2;

	if (parse_config ())		/* should fill envdevices */
		return 1;

	if ((addr1 = calloc (1, ENV_SIZE)) == NULL) {
		fprintf (stderr,
			"Not enough memory for environment (%ld bytes)\n",
			ENV_SIZE);
		return (errno);
	}

	/* read environment from FLASH to local buffer */
	environment.data = addr1;
	if (flash_io (O_RDONLY)) {
		return (errno);
	}

	crc1_ok = ((crc1 = crc32 (0, environment.data, env_size))
			   == environment.crc);
	flag1 = environment.flags;

	environment.data = addr1;
	environment.flags = flag1;
	environment.crc = crc1;
	return (0);
}


static int parse_config ()
{
	struct stat st;

	if (stat (device, &st)) {
		fprintf (stderr,
			"Cannot access MTD device %s: %s\n",
			device, strerror (errno));
		return 1;
	}

	return 0;
}

